#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#ifdef ULTRIX
  #include <sys/types.h>
#endif

/*#define DEBUG 1*/

void qualifyurl (char *url, char *dest);
void localname (char *url, char *dest);
void relativename (char *t1, char *dest);

/*
chaserefs v1.4 [part of the snag-o-rama distribution]

Copyright (c) 1996 by Joseph Clancy (jclancy@wimsey.com)
Placed under the GPL; see COPYING for further info.
*/




FILE *cmd, *info;
char id[500];
char urlparent[500];
char urlorig[500];
char urlhost[100];
char urlbase[500];
char urlpath[500];
char options[500];
char xfercmd[100];
int verbose, noimages, chaseparents;
int imgref, htmlref;
int imgrefchase, htmlrefchase;
long charcount, imgsize;
int depth;


void condenseurl (char *url)
{
	int i, j;
	char temp1[1000];
	char temp2[1000];

	strcpy (temp1, url);

	for (i = 0; i < strlen (temp1) - 1; i ++)
		if ((temp1[i] == '.') && (temp1[i + 1] == '.')) {
			if (i == 0)
				break;

			strcpy (temp2, temp1);

			for (j = i - 2; j >= 0; j --) {
				if (temp1[j] == '/')
					break;
			}

			strcpy (temp1 + j, temp2 + i + 2);
			i = 0;
		}

	strcpy (url, temp1);
	for (i = 0; i < strlen (temp1) - 2; i ++) {
		if ((temp1[i] == '/') && (temp1[i + 1] == '.') && 
(temp1[i + 2] != '\0')) {
			strcpy (url + i, temp1 + i + 2);
			strcpy (temp1, url);
			i = 0;
		}
	}
}



/*  Get the hostname (and port #) of a url  */

void gethstname (char *url, char *dest)
{
	int i;

	for (i = 7; i <= strlen (url); i ++) {
		if (url[i] == '/') {
			dest[i - 7] = '\0';
			break;
		}
		dest[i - 7] = url[i];
	}
}

/*  Decide if url refers to a file and is NOT a directory that we'll
    have to add index.html to  */

int isfile (char *url)
{
	int i, start;

	for (i = 7; i < strlen (url); i ++) {
		if (url[i] == '/')
			break;
		if (i == strlen (url))  /*  just a hostname  */
			return (0);
	}

	start = i + 1;

	for (i = strlen (url) - 1; i >= start; i --) {
		if (url[i] == '/')
			return (0);
		if (url[i] == '-')
			continue;
		if (url[i] == '_')
			continue;
		if (url[i] == '~')
			continue;
		if (!isalnum (url[i]))
			return (1);
	} 

	return (0);
}


int islocal (char *url)
{
	return (!strncmp (url + 7, urlhost, strlen (urlhost)));
}


/*  Get the path portion of a url, less the trailing filename.  */

void getpath (char *url, char *dest)
{
	int i;
	char temp1[1000];
	char temp2[1000];

	dest[0] = '\0';

	gethstname (url, temp2);

	if (strlen (url) - strlen (temp1) <= 2)
		return;

	strcpy (temp1, url + 8 + strlen (temp2));

	if (isfile (url)) {
		for (i = strlen (temp1) - 1; i > 0; i --)
			if (temp1[i] == '/')
				break;
		temp1[i] = '\0';
	} else
		if (temp1[strlen (temp1) - 1] == '/')
			temp1[strlen (temp1) - 1] = '\0';
 
	strcpy (dest, temp1); 
}


void getbase (char *url, char *dest)
{
	int i;
	char temp1[1000];
  
	if (!isfile (url)) {
		strcpy (dest, "index.html");
		return;
	}

	strcpy (temp1, url);

	for (i = 0; i < strlen (url); i ++)
		if (temp1[i] == '#')
			temp1[i] = '\0';

	for (i = strlen (temp1) - 1; i >= 0; i --)
		if (url[i] == '/') {
			strcpy (dest, url + i + 1);
			return;
		}

	strcpy (dest, "index.html");
}



/*  Assume pth to be ok - no funky characters  */

void ensurepath (char *pth, int includesbase)
{
	int i;
	char temp1[1000];
	struct stat test;

	if (pth[0] == '\0')
		return;

	strcpy (temp1, pth);
	if (!includesbase)
		strcat (temp1, "/");

	for (i = 0; i < strlen (temp1); i ++) {
		if (temp1[i] == '/') {
			temp1[i] = '\0';

			if ((stat (temp1, &test) == -1) && (errno == ENOENT))
				mkdir (temp1, 0777);

			temp1[i] = '/';
		}
	}
}


long filesize (char *fname)
{
	struct stat test;

	if (stat (fname, &test) == -1)
		return (-1);

	return (test.st_size);
}



int fileexists (char *fname)
{
	struct stat test;

	return (!((stat (fname, &test) == -1) && (errno == ENOENT)));
}



void printbanner ()
{
	fprintf (stderr, "chaserefs v1.4 (verbose commentary");

	if (chaseparents)
		fprintf (stderr, ", chasing parents");
	if (noimages)
		fprintf (stderr, ", no images");      
	
	fprintf (stderr, ")\n  command script: %s\n", id);
	fprintf (stderr, "  using %s to download over http\n", xfercmd);
	fprintf (stderr, "  url: %s\n\n", urlorig);  
}



void dashslashes (char *url)
{
	int i;

	for (i = 0; i < strlen (url); i ++)
		if (url[i] == '/')
			url[i] = '-';
}



void sterilizepath (char *pth)
{
	int i;
	char temp1[1000];

	strcpy (temp1, pth);

	for (i = 0; i < strlen (temp1); i ++)
		if (!isalnum (temp1[i]))
			switch (temp1[i]) {
			case '~':
				temp1[i] = '_';
				break;

			case '#':
			case '<':
			case '>':
			case ' ':
				temp1[i] = '\0';
				break;

			case '/':
				if (temp1[i + 1] == '/')
					temp1[i] = '-'; 
			case '.':
			case '_':
				break;

			case '?':
			case '&':
				dashslashes (temp1 + i);
			default:
				temp1[i] = '-';
				break;
			}

	strcpy (pth, temp1);
}



/*  Get command line arguments:
    argv[0] <url> <id> <level> [flags] ... */

void getargs (int argc, char **argv)
{
	int i, level;
	char temp1[1000];

	if (argc < 4) {
		fprintf (stderr, "chaserefs.c: getargs (aack)\n");
		fprintf (stderr, "You should probably be calling chaserefs from within\n");
		fprintf (stderr, "the snag/snagit shell scripts, unless you really know\n");
		fprintf (stderr, "what you're doing.\n\n");
		exit (1);
	}

	/*  Initialize globals  */

	verbose = noimages = chaseparents = 0;
	imgref = imgrefchase = 0;
	htmlref = htmlrefchase = 0;
	charcount = imgsize = 0;

	/*  Level of recursion since first snag call  */
	level = atoi (argv[3]);
	sprintf (options, "%d", level + 1);

	for (i = 4; i < argc; i ++) {
		strcat (options, " ");
		strcat (options, argv[i]);
		if (!strcmp (argv[i], "verbose"))
			verbose = 1;
		if (!strcmp (argv[i], "chaseparents"))
			chaseparents = 1;
		if (!strcmp (argv[i], "noimages"))
			noimages = 1;
	}

	strcpy (urlorig, argv[1]);
	if (strncmp (urlorig, "http://", 7)) {
		fprintf (stderr, "chaserefs.c: getargs (aack)\n");
		fprintf (stderr, "Hmm... the specified url doesn't seem to start with\n");
		fprintf (stderr, "http://.  We assume that it does...\n\n");
		exit (1);
	}

	strcpy (id, argv[2]);

	gethstname (urlorig, urlhost);
	getpath (urlorig, urlpath);
	getbase (urlorig, urlbase);

/*
	fprintf (stderr, "chaserefs: hst/path/base - %s %s %s\n", urlhost, 
urlpath, urlbase);
*/

	depth = (urlpath[0] == '\0' ? 0 : 1);
	for (i = 0; i < strlen (urlpath); i ++)
		if (urlpath[i] == '/')
			depth ++;

	strcpy (urlparent, getenv ("URL"));
	if (getenv ("XFERCMD") == NULL)
		strcpy (xfercmd, "webget -q");
	else
		strcpy (xfercmd, getenv ("XFERCMD"));

	if (verbose)
	{
		if (!level)
			printbanner ();

		fprintf (stderr, "%s/%s [%d]\n", urlpath, urlbase, level);
	}

	cmd = fopen (id, "at");
	if (verbose)
		fprintf (cmd, "\n\n#\n# chaserefs %s\n#\n\n", urlorig);

	getpath (urlorig, temp1);
	sterilizepath (temp1);
	ensurepath (temp1, 0);
}



void cleanup ()
{
  fclose (cmd);
  if (verbose)
  {
    fprintf (stderr, "%s/%s: %ld bytes\n", urlpath, urlbase, charcount);
    fprintf (stderr, "* downloaded %ldk by chasing %d of %d inlines and image refs\n", 
        imgsize / 1024, imgrefchase, imgref);
    fprintf (stderr, "* will chase %d of %d html refs\n", htmlrefchase, 
        htmlref);
    fprintf (stderr, "\n");
  }
  exit (0);
}


/*  Report if url is a local file reference ("path/file.html") or
    is instead fully specified ("http://..." or "ftp://...")  */

int relativeurl (char *url)
{
	int i;

	for (i = 0; ((i < strlen (url)) && (i < 9)); i ++) {
		if (url[i] == ':')
			return (0);
		if (!isalpha (url[i]))
			return (1);
	}

	return (1);
}



void qualifyurl (char *url, char *dest)
{
	int i;

	if (relativeurl (url)) {
		strcpy (dest, "http://");
		strcat (dest, urlhost);
		if (url[0] == '/')
			strcat (dest, url);
		else {
			if (urlpath[0] != '\0') {
				strcat (dest, "/");
				strcat (dest, urlpath);
			}
			strcat (dest, "/");
			strcat (dest, url);
		}
	} else {
		strcpy (dest, url);

		/*  might be a ftp/gopher/finger thing; won't touch  */
		if (strncmp (url, "http://", 7)) 
			return;
	}
	
	/*  now let's add index.html if necessary  */
	if (!isfile (dest))
		if (dest[strlen (dest) - 1] == '/')
			strcat (dest, "index.html");
		else
			strcat (dest, "/index.html");
}


/*  Converts references like http://aack/foo.html?hello to
    http://aack/foo-hello.html.  */

void patchurl (char *url)
{
	int i, j;
	char temp1[500];
	char ext[10];

	ext[0] = '\0';

	for (i = 0; i <= strlen (url); i ++) {
		if ((url[i] == '&') || (url[i] == '?'))
			for (j = 5; j; j --)
				if (url[j - 5] == '.') {
					memcpy (temp1 + i - j, url + i, 
strlen (url + i) + 1);
					memcpy (ext, url + i - j, j);
					ext[j] = '\0';
					strcat (temp1, ext);
					strcpy (url, temp1);
					return;
				}
		temp1[i] = url[i];
	}
}



void localname (char *url, char *dest)
{
	char temp1[1000];
	char origwd[1000];

	if (strncmp (url, "http://", 7)) {
		strcpy (dest, url);
		return;
	}

	/*  bug -> http://vc and http://vcc are seen as same  */
	if (strncmp (url + 7, urlhost, strlen (urlhost))) {
		strcpy (temp1, "/hosts/");
		strcat (temp1, url + 7);
	} else
		strcpy (temp1, url + 7 + strlen (urlhost));

	patchurl (temp1);
	sterilizepath (temp1);

	strcpy (dest, temp1 + 1);
}



void relativename (char *t1, char *dest)
{
	int i;
	char *owd;
	char temp1[1000];
	char temp2[1000];
	char origwd[1000];

	getpath (urlorig, origwd);
	owd = origwd;

	for (i = 0; i < strlen (owd); i ++) {
		if (owd[i] != t1[i])
			break;
		if (owd[i] == '/') {
			owd += (i + 1);
			t1 += (i + 1);
			i = -1;
			continue;
		}		
		if (owd[i + 1] == '\0') {
			owd += (i + 1);
			t1 += (i + 2);
			break;
		}
	}

	strcpy (temp1, t1);

	for (i = 0; i < strlen (owd); i ++)
		if (owd[i] == '/') {
			strcpy (temp2, "../");
			strcat (temp2, temp1);
			strcpy (temp1, temp2);
			owd += (i + 1);
			i = -1;
		}
	
	if (i) {
		strcpy (dest, "../");
		strcat (dest, temp1);
	} else
		strcpy (dest, temp1);
}



int ishtml (char *url)
{
	int i;
	char temp1[1000];

	strcpy (temp1, url);

	for (i = 0; i < strlen (temp1); i ++)
		temp1[i] = tolower (temp1[i]);

	if (!strcmp (temp1 + strlen (temp1) - 4, ".htm"))
		return (1);
	if (!strcmp (temp1 + strlen (temp1) - 5, ".html"))
		return (1);
	if (!strcmp (temp1 + strlen (temp1) - 6, ".shtml"))
		return (1);
	if (!strcmp (temp1 + strlen (temp1) - 3, ".pl"))
		return (1);
	if (!strcmp (temp1 + strlen (temp1) - 4, ".cgi"))
		return (1);

	return (0);
}


int ishttp (char *data)
{
	int i;
	char temp1[10];

	for (i = 0; i < 7; i ++)
		temp1[i] = tolower (data[i]);

	temp1[7] = '\0';

	return (!strcmp (temp1, "http://"));
}



int dontwant (char *nm)
{
	int i;
	char temp1[1000];

	strcpy (temp1, nm);

	for (i = 0; i < strlen (temp1); i ++)
		temp1[i] = tolower (temp1[i]);

	if (!strcmp (temp1 + strlen (temp1) - 4, ".gif"))
		return (noimages);
	if (!strcmp (temp1 + strlen (temp1) - 4, ".xbm"))
		return (noimages);
	if (!strcmp (temp1 + strlen (temp1) - 4, ".jpg"))
		return (noimages);
	if (!strcmp (temp1 + strlen (temp1) - 5, ".jpeg"))
		return (noimages);
	if (!strcmp (temp1 + strlen (temp1) - 5, ".html"))
		return (0);
	if (!strcmp (temp1 + strlen (temp1) - 4, ".htm"))
		return (0);
	if (!strcmp (temp1 + strlen (temp1) - 3, ".pl"))
		return (0);
	if (!strcmp (temp1 + strlen (temp1) - 4, ".cgi"))
		return (0);
	if (!strcmp (temp1 + strlen (temp1) - 3, ".sh"))
		return (0);

/*  should depend on command line flag: anyfiles?  */
	return (noimages);
}



int isevilparent (char *url)
{
	char host[500];
	char path[500];

	if (chaseparents)
		return (0);

	gethstname (urlparent, host);
	getpath (urlparent, path);

	return (memcmp (url, urlparent, strlen (path) + 7 + strlen (host)));
}



void handlename (char *data)
{
	int i, parent;
	char url[1000];
	char locname[1000];
	char relname[1000];
	char temp1[1000];
	long size;

	/*  internal named href  */
	if (data[0] == '#') {
		printf ("\"%s\"", data);
		return;
	}

	qualifyurl (data, url);
	condenseurl (url);
	localname (url, locname);
	condenseurl (locname);
	relativename (locname, relname);

#ifdef DEBUG
	fprintf (stderr, "\n%s qualified as %s saved to %s\n", data, url, locname);
#endif

	if ((!ishttp (url)) || (dontwant (locname))) {
		printf ("\"%s\"", data);
		return;
	}

	if (!islocal (url)) {  
		printf ("\"%s\"", data);
		return;
	}

	parent = isevilparent (url);

	if ((ishtml (locname)) && (!parent)) {
		htmlref ++;
		ensurepath (locname, 1);
		sprintf (temp1, "%s.real", locname);
		if (!fileexists (temp1)) {
			htmlrefchase ++;
			fprintf (cmd, "snagit \"%s\" %s %s %s\n", url, 
locname, id, options);
			if (verbose)
				fprintf (stderr, "  -> %s\n", locname);
		}
	} else if (!ishtml (locname)) {
		imgref ++;
		ensurepath (locname, 1);
		if (!fileexists (locname)) {
			if (verbose)
				fprintf (stderr, "     %s", locname);
			sprintf (temp1, "%s \"%s\" > \"%s\"", xfercmd, url, locname);
			system (temp1);
			size = filesize (locname);
			if (verbose)
				fprintf (stderr, " (%ldk)\n", size / 1024);
		        imgrefchase ++;
			imgsize += size;
		}
	}

	printf ("\"%s\"", relname);
}



int okcontext (char *tag)
{
	if (!strcmp (tag, "href"))
		return (1);
	if (!strcmp (tag, "src"))
		return (1);
	if (!strcmp (tag, "background"))
		return (1);

	return (0);
}

void filter ()
{
  int tag, name, word, i;
  char c;
  char data[1000];
  char wdata[1000];
  char tdata1[10];
  char tdata2[10];
  char model[10];

  tag = name = word = 0;
  data[0] = wdata[0] = '\0';

  while (1) {
    c = getc (stdin);
    if (feof (stdin))
      break;
/*
#ifdef DEBUG
	fprintf (stderr, "%c", c);
#endif
*/
    charcount ++;
    if (name) {
      if (c == '"') {
        name = 0;
	if (okcontext (wdata))
		handlename (data);
	else
		printf ("\"%s\"", data);
	data[0] = '\0';
        continue;
      }
      data[strlen (data) + 1] = '\0';
      data[strlen (data)] = c;
      continue;
    }
    if (tag) {
      if (c == '>')
        tag = 0;
      if (isalpha (c)) {
        if (!word) {
          word = 1;
          wdata[0] = '\0';
        }
        wdata[strlen (wdata) + 1] = '\0';
        wdata[strlen (wdata)] = tolower (c);
      }
      else if (word) 
        word = 0;
      if (c == '"') {
        name = 1;
        continue;
      }
    }
    else if (c == '<')
      tag = 1;
    putc (c, stdout);
  }  
}


void moveindex ()
{
  char dest[500];
  int i;

  strcpy (dest, urlbase);
  if (urlpath[0] == '\0')
    if (strcmp (urlbase, "index.html"))
    {
      fprintf (cmd, "\n# initial snagged file\nmv index.html %s\n", dest);
      fprintf (cmd, "mv index.html.real %s.real\n", dest);
    }
    else
      return;
  sprintf (dest, "%s/%s", urlpath, urlbase);
  for (i = 0; i < strlen (dest); i ++)
    if (dest[i] == '~')
      dest[i] = '_';
  fprintf (cmd, "\n# initial snagged file\nmv index.html %s\n", dest);
  fprintf (cmd, "mv index.html.real %s.real\n", dest);
}

int main (int argc, char **argv)
{
  getargs (argc, argv);
  filter (); 
  if (atoi (argv[3]) == 0)
    moveindex ();
  cleanup ();
  return (0);
}


